@@ -4,7 +4,7 @@ RDF::Endpoint - A SPARQL Protocol Endpoint implementation
 
 =head1 VERSION
 
-This document describes RDF::Endpoint version 0.05.
+This document describes RDF::Endpoint version 0.07.
 
 =head1 SYNOPSIS
 
@@ -103,15 +103,16 @@ package RDF::Endpoint;
 use 5.008;
 use strict;
 use warnings;
-our $VERSION	= '0.05';
+our $VERSION	= '0.07';
 
 use RDF::Query 2.905;
 use RDF::Trine 0.134 qw(statement iri blank literal);
 
+use JSON;
 use Encode;
 use File::Spec;
 use Data::Dumper;
-use Digest::MD5 qw(md5_hex);
+use Digest::MD5 qw(md5_base64);
 use XML::LibXML 1.70;
 use Plack::Request;
 use Plack::Response;
@@ -139,6 +140,8 @@ my $NAMESPACES	= {
 	ldodds		=> 'java:com.ldodds.sparql.',
 	fn			=> 'http://www.w3.org/2005/xpath-functions#',
 	sparql		=> 'http://www.w3.org/ns/sparql#',
+	vann		=> 'http://purl.org/vocab/vann/',
+	sde			=> 'http://kasei.us/ns/service-description-extension#',
 };
 
 =item C<< new ( \%conf ) >>
@@ -161,12 +164,25 @@ sub new {
 	} else {
 		$config		= $arg;
 		my $store	= RDF::Trine::Store->new( $config->{store} );
+		unless ($store) {
+			warn "Failed to construct RDF Store object";
+			return;
+		}
 		$model		= RDF::Trine::Model->new( $store );
+		unless ($model) {
+			warn "Failed to construct RDF Model object";
+			return;
+		}
 	}
 	
 	unless ($config->{endpoint}) {
 		$config->{endpoint}	= { %$config };
 	}
+	
+	if ($config->{endpoint}{load_data} and $config->{endpoint}{update}) {
+		die "The load_data and update configuration options cannot be specified together.";
+	}
+	
 	my $self	= bless( {
 		conf		=> $config,
 		model		=> $model,
@@ -194,6 +210,11 @@ sub run {
 	
 	my $content;
 	my $response	= Plack::Response->new;
+
+	my $server = "RDF::Endpoint/$VERSION";
+	$server .= " " . $response->headers->header('Server') if defined($response->headers->header('Server'));
+	$response->headers->header('Server' => $server);
+
 	unless ($req->path eq $endpoint_path) {
 		my $path	= $req->path_info;
 		$path		=~ s#^/##;
@@ -230,9 +251,11 @@ END
 		my $method	= uc($req->method);
 		$content	= "Unexpected method $method (expecting GET or POST)";
 		$self->log_error( $req, $content );
-		my $code	= 405;
-		$response->status("$code Method Not Allowed");
-		$response->body($content);
+		$self->_set_response_error($req, $response, 405, {
+			title		=> 'Method not allowed',
+			describedby	=> 'http://id.kasei.us/perl/rdf-endpoint/error/bad_http_method',
+		});
+		$response->header('Allow' => 'GET, POST');
 		goto CLEANUP;
 	} elsif (defined($ct) and $ct eq 'application/sparql-query') {
 		$sparql	= $req->content;
@@ -245,9 +268,10 @@ END
 		if (scalar(@sparql) > 1) {
 			$content	= "More than one query string submitted";
 			$self->log_error( $req, $content );
-			my $code	= 400;
-			$response->status("$code Multiple Query Strings");
-			$response->body($content);
+			$self->_set_response_error($req, $response, 400, {
+				title		=> 'Multiple query strings not allowed',
+				describedby	=> 'http://id.kasei.us/perl/rdf-endpoint/error/multiple_queries',
+			});
 			goto CLEANUP;
 		} else {
 			$sparql = $sparql[0];
@@ -257,9 +281,10 @@ END
 		if (scalar(@sparql) > 1) {
 			$content	= "More than one update string submitted";
 			$self->log_error( $req, $content );
-			my $code	= 400;
-			$response->status("$code Multiple Update Strings");
-			$response->body($content);
+			$self->_set_response_error($req, $response, 400, {
+				title		=> 'Multiple update strings not allowed',
+				describedby	=> 'http://id.kasei.us/perl/rdf-endpoint/error/multiple_updates',
+			});
 			goto CLEANUP;
 		}
 		
@@ -269,9 +294,11 @@ END
 			my $method	= $req->method;
 			$content	= "Update operations must use POST";
 			$self->log_error( $req, $content );
-			my $code	= 405;
-			$response->status("$code $method Not Allowed for Update Operation");
-			$response->body($content);
+			$self->_set_response_error($req, $response, 405, {
+				title		=> "$method Not Allowed for Update Operation",
+				describedby	=> 'http://id.kasei.us/perl/rdf-endpoint/error/bad_http_method_update',
+			});
+			$response->header('Allow' => 'POST');
 			goto CLEANUP;
 		}
 	}
@@ -283,21 +310,34 @@ END
 		$args{ update }		= 1 if ($config->{endpoint}{update} and $req->method eq 'POST');
 		$args{ load_data }	= 1 if ($config->{endpoint}{load_data});
 		
-		my @default	= $req->param('default-graph-uri');
-		my @named	= $req->param('named-graph-uri');
-		if (scalar(@default) or scalar(@named)) {
-			delete $args{ load_data };
-			$model	= RDF::Trine::Model->new( RDF::Trine::Store::Memory->new() );
-			foreach my $url (@named) {
-				RDF::Trine::Parser->parse_url_into_model( $url, $model, context => iri($url) );
+		{
+			my @default	= $req->param('default-graph-uri');
+			my @named	= $req->param('named-graph-uri');
+			if (scalar(@default) or scalar(@named)) {
+				delete $args{ load_data };
+				$model	= RDF::Trine::Model->new( RDF::Trine::Store::Memory->new() );
+				foreach my $url (@named) {
+					RDF::Trine::Parser->parse_url_into_model( $url, $model, context => iri($url) );
+				}
+				foreach my $url (@default) {
+					RDF::Trine::Parser->parse_url_into_model( $url, $model );
+				}
 			}
-			foreach my $url (@default) {
-				RDF::Trine::Parser->parse_url_into_model( $url, $model );
+		}
+		
+		my $protocol_specifies_update_dataset	= 0;
+		{
+			my @default	= $req->param('using-graph-uri');
+			my @named	= $req->param('using-named-graph-uri');
+			if (scalar(@named) or scalar(@default)) {
+				$protocol_specifies_update_dataset	= 1;
+				$model	= RDF::Trine::Model::Dataset->new( $model );
+				$model->push_dataset( default => \@default, named => \@named );
 			}
 		}
 		
 		my $match	= $headers->header('if-none-match') || '';
-		my $etag	= md5_hex( join('#', $self->run_tag, $model->etag, $type, $ae, $sparql) );
+		my $etag	= md5_base64( join('#', $self->run_tag, $model->etag, $type, $ae, $sparql) );
 		if (length($match)) {
 			if (defined($etag) and ($etag eq $match)) {
 				$response->status(304);
@@ -309,13 +349,31 @@ END
 		my $query	= RDF::Query->new( $sparql, { lang => 'sparql11', base => $base, %args } );
 		$self->log_query( $req, $sparql );
 		if ($query) {
+			if ($protocol_specifies_update_dataset and $query->specifies_update_dataset) {
+				my $method	= $req->method;
+				$content	= "Update operations cannot specify a dataset in both the query and with protocol parameters";
+				$self->log_error( $req, $content );
+				$self->_set_response_error($req, $response, 400, {
+					title		=> "Multiple datasets specified for update",
+					describedby	=> 'http://id.kasei.us/perl/rdf-endpoint/error/update_specifies_multiple_datasets',
+					detail		=> $content,
+				});
+				goto CLEANUP;
+			}
 			my ($plan, $ctx)	= $query->prepare( $model );
 # 			warn $plan->sse;
 			my $iter	= $query->execute_plan( $plan, $ctx );
 			if ($iter) {
 				$response->status(200);
 				if (defined($etag)) {
-					$response->headers->header( ETag => $etag );
+					if ($etag !~ /"/) {
+						$etag	= qq["$etag"];
+					}
+					if ($etag =~ qr[^(W/)?"[\x{21}\x{23}-\x{7e}\x{80}-\x{FF}]*"$]) {
+						$response->headers->header( ETag => $etag );
+					} else {
+						warn "ETag value is not syntactically valid: " . Dumper($etag);
+					}
 				}
 				if ($iter->isa('RDF::Trine::Iterator::Graph')) {
 					my @variants	= (['text/html', 0.99, 'text/html']);
@@ -368,28 +426,38 @@ END
 				}
 			} else {
 				my $error	= $query->error;
-				$self->log_error( $req, "$error\t$sparql" );
-				$response->status(500);
-				$response->body($query->error);
+				$self->_set_response_error($req, $response, 500, {
+					title		=> "SPARQL query/update execution error",
+					describedby	=> 'http://id.kasei.us/perl/rdf-endpoint/error/execution_error',
+					detail		=> "$error; $sparql",
+				});
 				$content	= RDF::Query->error;
 			}
 		} else {
 			$content	= RDF::Query->error;
 			$self->log_error( $req, $content );
 			my $code	= ($content =~ /Syntax/) ? 400 : 500;
-			my $message	= ($code == 400) ? "Syntax Error" : "Internal Server Error";
-			$response->status("$code $message");
-			$response->body($content);
 			if ($req->method ne 'POST' and $content =~ /read-only queries/sm) {
 				$content	= 'Updates must use a HTTP POST request.';
+				$self->_set_response_error($req, $response, $code, {
+					title		=> $content,
+					describedby	=> 'http://id.kasei.us/perl/rdf-endpoint/error/bad_http_method_update',
+				});
+			} else {
+				$self->_set_response_error($req, $response, $code, {
+					title		=> "SPARQL query/update parse error",
+					describedby	=> 'http://id.kasei.us/perl/rdf-endpoint/error/parse_error',
+					detail		=> $content,
+				});
 			}
 		}
 	} elsif ($req->method eq 'POST') {
 		$content	= "POST without recognized query or update";
 		$self->log_error( $req, $content );
-		my $code	= 400;
-		$response->status("$code Missing SPARQL Query/Update String");
-		$response->body($content);
+		$self->_set_response_error($req, $response, 400, {
+			title		=> "Missing SPARQL Query/Update String",
+			describedby	=> 'http://id.kasei.us/perl/rdf-endpoint/error/missing_sparql_string',
+		});
 	} else {
 		my @variants;
 		my %media_types	= %RDF::Trine::Serializer::media_types;
@@ -421,6 +489,9 @@ END
 	}
 	
 CLEANUP:
+# 	warn Dumper($model);
+# 	warn $model->as_string;
+	$content	= $response->body || $content;
 	my $length	= 0;
 	my %ae		= map { $_ => 1 } split(/\s*,\s*/, $ae);
 	if ($ae{'gzip'}) {
@@ -456,7 +527,7 @@ Returns a unique key for each instantiation of this service.
 
 sub run_tag {
 	my $self	= shift;
-	return md5_hex(refaddr($self) . $self->{start_time});
+	return md5_base64(refaddr($self) . $self->{start_time});
 }
 
 =item C<< service_description ( $request, $model ) >>
@@ -479,7 +550,10 @@ sub service_description {
 	}
 	
 	my $config		= $self->{conf};
+	my $doap		= RDF::Trine::Namespace->new('http://usefulinc.com/ns/doap#');
 	my $sd			= RDF::Trine::Namespace->new('http://www.w3.org/ns/sparql-service-description#');
+	my $sde			= RDF::Trine::Namespace->new('http://kasei.us/ns/service-description-extension#');
+	my $vann		= RDF::Trine::Namespace->new('http://purl.org/vocab/vann/');
 	my $void		= RDF::Trine::Namespace->new('http://rdfs.org/ns/void#');
 	my $scovo		= RDF::Trine::Namespace->new('http://purl.org/NET/scovo#');
 	my $count		= $model->count_statements( undef, undef, undef, RDF::Trine::Node::Nil->new );
@@ -539,6 +613,25 @@ sub service_description {
 		}
 	}
 	
+	if (my $software = $config->{endpoint}{service_description}{software}) {
+		$sdmodel->add_statement( statement( $s, $sde->software, iri($software) ) );
+	}
+	
+	if (my $related = $config->{endpoint}{service_description}{related}) {
+		foreach my $r (@$related) {
+			$sdmodel->add_statement( statement( $s, $sde->relatedEndpoint, iri($r) ) );
+		}
+	}
+	
+	if (my $namespaces = $config->{endpoint}{service_description}{namespaces}) {
+		while (my($ns,$uri) = each(%$namespaces)) {
+			my $b	= RDF::Trine::Node::Blank->new();
+			$sdmodel->add_statement( statement( $s, $sde->namespace, $b ) );
+			$sdmodel->add_statement( statement( $b, $vann->preferredNamespacePrefix, literal($ns) ) );
+			$sdmodel->add_statement( statement( $b, $vann->preferredNamespaceUri, literal($uri) ) );
+		}
+	}
+	
 	$self->{ sd_cache }	= [ $etag, $sdmodel ];
 	return $sdmodel;
 }
@@ -734,6 +827,33 @@ sub _log {
 	$logger->($data);
 }
 
+sub _set_response_error {
+	my $self	= shift;
+	my $req		= shift;
+	my $resp	= shift;
+	my $code	= shift;
+	my $error	= shift;
+	my @variants	= (
+		['text/plain', 1.0, 'text/plain'],
+		['application/json-problem', 0.99, 'application/json-problem'],
+	);
+	my $headers	= $req->headers;
+	my $stype	= choose( \@variants, $headers ) || 'text/plain';
+	if ($stype eq 'application/json-problem') {
+		$resp->headers->content_type( 'application/json-problem' );
+		$resp->status($code);
+		my $content	= encode_json($error);
+		$resp->body($content);
+	} else {
+		$resp->headers->content_type( 'text/plain' );
+		$resp->status($code);
+		my @messages	= grep { defined($_) } @{ $error }{ qw(title detail) };
+		my $content		= join("\n\n", @messages);
+		$resp->body($content);
+	}
+	return;
+}
+
 =end private
 
 =cut
@@ -764,7 +884,7 @@ __END__
 
 =head1 LICENSE AND COPYRIGHT
 
-Copyright (c) 2010-2012 Gregory Todd Williams.
+Copyright (c) 2010-2014 Gregory Todd Williams.
 
 This software is provided 'as-is', without any express or implied
 warranty. In no event will the authors be held liable for any
